/**
 * @module
 *
 * The complete Discourse API, fully typed.
 *
 * @example
 * ```ts
 * import Discourse from "discourse2";
 *
 * const discourse = new Discourse("https://forums.kiri.art", {
 *   "Api-Key": process.env.DISCOURSE_API_KEY,
 *   "Api-Username": process.env.DISCOURSE_API_USERNAME,
 * });
 *
 * const result = await discourse.listLatestTopics();
 * ```
 */
import { Ajv } from "ajv";
import DiscourseAPIGenerated from "./generated.js";
type Prettify<T> = {
    [K in keyof T]: T[K];
} & {};
declare const ajv: Ajv;
/**
 * Returned when the HTTP status code is not 200.
 * The error message is available as the `message` property.
 * The status code is available as the `status` property.
 * The full response is available in the `request` property.
 */
export declare class HTTPError extends Error {
    status: number;
    response: Response;
    constructor(status: number, message: string, response: Response);
}
/**
 * Returned when the parameters provided to a method are invalid.
 * The exact (AJV) errors are available in the `errors` property.
 */
export declare class ParamaterValidationError extends Error {
    errors: typeof ajv.errors;
    constructor(message: string, errors: typeof ajv.errors);
}
/**
 * Returned when the response from the server is invalid.
 * The exact (AJV) errors are available in the `errors` property.
 */
export declare class ResponseValidationError extends Error {
    errors: typeof ajv.errors;
    constructor(message: string, errors: typeof ajv.errors);
}
/**
 * The full Discourse API class.
 *
 * @example
 * ```ts
 * import Discourse from "discourse2";
 *
 * const discourse = new Discourse("https://forums.kiri.art", {
 *   "Api-Key": process.env.DISCOURSE_API_KEY,
 *   "Api-Username": process.env.DISCOURSE_API_USERNAME,
 * });
 *
 * const result = await discourse.listLatestTopics();
 * ```
 */
export default class DiscourseAPI extends DiscourseAPIGenerated {
    private url;
    private "Api-Key"?;
    private "Api-Username"?;
    constructor(url: string, opts?: {
        "Api-Key"?: string;
        "Api-Username"?: string;
    });
    _exec<T>(operationName: string, params?: Record<string, string>): Promise<unknown>;
    createUpload(params: Prettify<Omit<Parameters<DiscourseAPIGenerated["createUpload"]>[0], "file"> & {
        file?: File | Blob;
    }>): Prettify<ReturnType<DiscourseAPIGenerated["createUpload"]>>;
    updateTopic(params: Parameters<DiscourseAPIGenerated["updateTopic"]>[0]): ReturnType<DiscourseAPIGenerated["updateTopic"]>;
    getTopicByExternalId(params: Parameters<DiscourseAPIGenerated["getTopicByExternalId"]>[0]): ReturnType<DiscourseAPIGenerated["getTopic"]>;
    getSpecificPostsFromTopic(params: Parameters<DiscourseAPIGenerated["getSpecificPostsFromTopic"]>[0]): ReturnType<DiscourseAPIGenerated["getSpecificPostsFromTopic"]>;
}
export {};
//# sourceMappingURL=index.d.ts.map